/*  easylog.h - a fairly minimal logging-to-file implementation

 Adrian Groves, FMRIB Image Analysis Group

 Copyright (C) 2007-2008 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#pragma once

#include <assert.h>
#include <iostream>
#include <map>
#include <sstream>
#include <stdexcept>
#include <string>
#include <vector>

/**
 * Use LOG just like you'd use cout.  (i.e. LOG << your_data << endl;)
 * Your class must be derived from Loggable.
 *
 * If the log member variable m_log is not set, log output goes to stderr
 */
#define LOG ((m_log == 0) ? std::cerr : m_log->LogStream())
#define LOG_ERR(x) ((void)((LOG << x)))

#define WARN_ONCE(x)                                                                               \
    if (m_log)                                                                                     \
    m_log->WarnOnce(x)
#define WARN_ALWAYS(x)                                                                             \
    if (m_log)                                                                                     \
    m_log->WarnAlways(x)

/**
 * Sends logging information to an output stream
 *
 * StartLog() is used to tell the logger where to put the output - any log
 * calls made before this are stored in a temporary stringstream and then
 * flushed to the log once StartLog is called.
 *
 * If you want to dump variables directly to this stream, just implement
 * std::ostream& operator<<(std::ostream& stream, const your_type& data);
 */
class EasyLog
{
public:
    /**
     * Log output to an existing stream
     *
     * This might be std::cout/cerr or something else, e.g. a
     * stringstream.
     */
    EasyLog();
    ~EasyLog();

    /**
     * Log output to a file
     *
     * The file will be called 'logfile'
     *
     * @param outDir Name of a directory in which to put the logfile.
     */
    void StartLog(const std::string &outDir);

    /**
     * Log output to an existing stream
     *
     * This might be std::cout/cerr or something else, e.g. a
     * stringstream.
     */
    void StartLog(std::ostream &s);

    /**
     * Get the output directory for the log file, or
     * an empty string if not logging to a file.
     */
    const std::string &GetOutputDirectory();

    /**
     * Stop logging.
     *
     * @param gzip If true and we are logging to a file, gzip the logfile
     */
    void StopLog(bool gzip = false);

    /**
     * @return true if StartLog has been called
     */
    bool LogStarted();

    /**
     * Get the logging stream. Easier to use the LOG macro defined above
     */
    std::ostream &LogStream();

    /**
     * Issue a warning
     *
     * The warning will appear once in the log.
     * If the same warning occurs again, it will not
     * be repeated, apart from if ReissueWarnings is
     * called
     */
    void WarnOnce(const std::string &text);

    /**
     * Issue a warning
     *
     * The warning will appear in the log each time it is issued.
     */
    void WarnAlways(const std::string &text);

    /**
     * Resend all warnings recorded so far to the log
     */
    void ReissueWarnings();

private:
    std::ostream *m_stream;
    std::stringstream m_templog;
    std::string m_outdir;
    std::map<std::string, int> m_warncount;
};

class Loggable
{
public:
    explicit Loggable(EasyLog *log = 0)
        : m_log(log)
        , m_debug(false)
    {
    }
    EasyLog *GetLogger() const
    {
        return m_log;
    }
    void SetLogger(EasyLog *log)
    {
        m_log = log;
    }

protected:
    EasyLog *m_log;
    bool m_debug;
};

/**
 * Convert almost anything to a string
 */
template <typename type> inline std::string stringify(type from)
{
    std::ostringstream s;
    if (!(s << from))
        throw std::logic_error("Stringify failed");
    return s.str();
}

/**
 * Output a vector<int>
 */
inline std::ostream &operator<<(std::ostream &out, std::vector<int> x)
{
    out << "[ ";
    for (unsigned i = 0; i < x.size(); i++)
        out << x[i] << " ";
    out << "]";
    return out;
}
