/*  fabber_capi.h - Pure C API for fabber

 Used by language bindings, e.g. Python interface

 Martin Craig

 Copyright (C) 2016 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#pragma once

// Export required symbols to Windows DLL
#ifdef _WIN32
#ifdef fabbercore_shared_EXPORTS
#define FABBER_DLL_API __declspec(dllexport)
#else
#define FABBER_DLL_API __declspec(dllimport)
#endif
#else
#define FABBER_DLL_API
#endif

#define FABBER_ERR_MAXC 255
#define FABBER_ERR_FATAL -255
#define FABBER_ERR_NEWMAT -254

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Create a new context for running fabber.
 *
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return A handle to the context. This should not be used for any purpose apart from
 *         to pass to other API functions
 */
FABBER_DLL_API void *fabber_new(char *err_buf);

/**
 * Load models from a dynamic library
 *
 * @param fab Fabber context, returned by fabber_new
 * @param libpath Path to library, non NULL, not empty
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_load_models(void *fab, const char *libpath, char *err_buf);

/**
 * Set the extent of the volume to be procesed.
 *
 * This must be called before any data is provided.
 *
 * @param fab Fabber context, returned by fabber_new
 * @param nx Extent in x direction
 * @param ny Extent in y direction
 * @param nz Extent in z direction
 * @param mask Array of length nx*ny*nz of 0 and 1 defining a mask region.
 *             (where mask=0, voxel is ignored, where mask!=0, voxel is included)
 *             Column-major order is used (see fabber_set_data). If NULL, data is
 *             assumed to be unmasked.
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_set_extent(
    void *fab, unsigned int nx, unsigned int ny, unsigned int nz, const int *mask, char *err_buf);

/**
 * Destroy fabber context previously created in fabber_new
 *
 * Will not return any errors.
 *
 * @param fab Fabber context, returned by fabber_new. NULL will be ignored. Anything
 *            else will probably cause a crash.
 */
FABBER_DLL_API void fabber_destroy(void *fab);

/**
 * Set an option
 *
 * @param fab Fabber context, returned by fabber_new
 * @param key Key value, not null or empty
 * @param value Value, not null, may be empty (e.g. for boolean options)
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_set_opt(void *fab, const char *key, const char *value, char *err_buf);

/**
 * Set voxel data
 *
 * @param fab Fabber context, returned by fabber_new
 * @param name Name of data. Main timeseries has name 'data'. Not null, not empty
 * @param data_size Data size in 4th dimension, i.e. number of t-slices for the main
 *                  timeseries, 1 for simple image data.
 * @param data Array of size nx*ny*nz*data_size. Will be masked using the mask supplied
 *             in fabber_new. Column-major order is used, i.e. 1st dimension varies fastest
 *             (default in FSL)
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_set_data(
    void *fab, const char *name, unsigned int data_size, const float *data, char *err_buf);

/**
 * Get the size of output voxel data
 *
 * @param fab Fabber context, returned by fabber_new
 * @param name Name of data. Not null, not empty
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return Size of data in 4th dimension, or <0 on error
 */
FABBER_DLL_API int fabber_get_data_size(void *fab, const char *name, char *err_buf);

/**
 * Get output voxel data
 *
 * @param fab Fabber context, returned by fabber_new
 * @param name Name of data. Not null, not empty
 * @param data_buf Data buffer of size nx*ny*nz*data_size. Data size may be obtained
 *                 from fabber_get_data_size. Column-major order is used (see fabber_set_data)
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_data(void *fab, const char *name, float *data_buf, char *err_buf);

/**
 * Run Fabber model fitting on already configured options and data
 *
 * @param fab Fabber context, returned by fabber_new
 * @param log_bufsize Size of the log buffer. If too small, only this number of characters
 *                    will be returned.
 * @param log_buf Char buffer of size log_bufsize to receive output log
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 * @param progress_cb Function pointer which takes two integers (current voxel, total voxels). Pass
 * NULL if not required
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_dorun(void *fab, unsigned int log_bufsize, char *log_buf, char *err_buf,
    void (*progress_cb)(int, int));

/**
 * Get fabber options, optionally for a specific method or model
 *
 * @param fab Fabber context, returned by fabber_new
 * @param key "method" for method option, "model" for model options, NULL or empty
 *            for general Fabber options
 * @param value Name of method or model. Ignored for general options
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size log_bufsize to receive output. First line will contain
 *                description of the model, method, or Fabber generally. This will be blank
 *                if no description exists, but the line will be included. The remaining lines
 *                will contain tab-separated data with each option on a single line. Order of
 *                data is option name, description, type, optional (1 or 0), default value
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_options(void *fab, const char *key, const char *value,
    unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Get list of known models
 *
 * @param fab Fabber context, returned by fabber_new
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size out_bufsize to receive output. Will contain
 *                known model names separated by newlines
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_models(
    void *fab, unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Get list of known inference methods
 *
 * @param fab Fabber context, returned by fabber_new
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size out_bufsize to receive output. Will contain
 *                known method names separated by newlines
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_methods(
    void *fab, unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Get a list of model parameters that will be output. Note that this will depend
 * on the options specified, so must be called after all options are set
 *
 * @param fab Fabber context, returned by fabber_new
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size out_bufsize to receive output. Will contain
 *                model parameter names separated by newlines
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_model_params(
    void *fab, unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Get a list of model parameters with descriptions. Note that this will depend
 * on the options specified, so must be called after all options are set
 *
 * @param fab Fabber context, returned by fabber_new
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size out_bufsize to receive output. Each line will
 *                contain parameter name followed by a space followed by an optinal 
 *                description string followed by the optional terminating string 
 *                (units: <units>)
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_model_param_descs(
    void *fab, unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Get a list of additional model-specific timeseries data that will be output. This
 * could depend on the options specified, so must be called after all options are set
 *
 * @param fab Fabber context, returned by fabber_new
 * @param out_bufsize Size of the output buffer. If too small, no output is returned
 * @param out_buf Char buffer of size out_bufsize to receive output. Will contain
 *                model output names separated by newlines
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_get_model_outputs(
    void *fab, unsigned int out_bufsize, char *out_buf, char *err_buf);

/**
 * Evaluate the model for a given set of parameters
 *
 * @param fab Fabber context, returned by fabber_new
 * @param n_params Number of parameters passed
 * @param params Array of parameter values in order returned by fabber_get_model_params
 * @param n_ts Length of input data / output buffer, i.e. number of t-points to evaluate
 * @param indata Array buffer for input data - may be NULL if not passing any input data
 * @param output Array buffer for output data
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_model_evaluate(void *fab, unsigned int n_params, float *params,
    unsigned int n_ts, float *indata, float *output, char *err_buf);

/**
 * Evaluate the model for a given set of parameters with the option of returning one
 * of the alternative model outputs (i.e. those returned by get_model_outputs).
 *
 * @param fab Fabber context, returned by fabber_new
 * @param n_params Number of parameters passed
 * @param params Array of parameter values in order returned by fabber_get_model_params
 * @param n_ts Length of input data / output buffer, i.e. number of t-points to evaluate
 * @param indata Array buffer for input data - may be NULL if not passing any input data
 * @param output_name Name of output to return - empty string for the default output
 * @param output Array buffer for output data
 * @param err_buf Optional buffer for error message. Max message length=FABBER_ERR_MAXC
 *
 * @return 0 on success, <0 on failure
 */
FABBER_DLL_API int fabber_model_evaluate_output(void *fab, unsigned int n_params, float *params, 
    unsigned int n_ts, float *indata, const char *output_name, float *output, char *err_buf);

#ifdef __cplusplus
}
#endif
