/*  inference.h - General inference technique base class

 Adrian Groves & Michael Chappell, FMRIB Image Analysis Group & IBME QuBIc Group

 Copyright (C) 2007-2015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#pragma once

#include "dist_mvn.h"
#include "easylog.h"
#include "factories.h"
#include "fwdmodel.h"
#include "noisemodel.h"
#include "rundata.h"

#include <iomanip>
#include <string>
#include <vector>

class InferenceTechnique : public Loggable
{
public:
    /**
     * Static member function to return the names of all known
     * inference techniques
     */
    static std::vector<std::string> GetKnown();

    /**
     * Static member function, to pick an inference technique from a name
     */
    static InferenceTechnique *NewFromName(const std::string &name);

    /**
     * Get usage information for a named method
     */
    static void UsageFromName(const std::string &name, std::ostream &stream);

    /**
     * Create a new instance of this class.
     * @return pointer to new instance.
     */
    static InferenceTechnique *NewInstance();

    /**
     * Default constructor.
     */
    InferenceTechnique();

    /**
     * Get option descriptions for this inference method.
     */
    virtual void GetOptions(std::vector<OptionSpec> &opts) const {};

    /**
     * @return human-readable description of the inference method.
     */
    virtual std::string GetDescription() const = 0;

    /**
     * Get the code version. There is no fixed format for this,
     * and it has no meaning other than by comparison with different
     * versions of the same inference method code.
     *
     * See fwdmodel.cc for an example of how to implement this to
     * return a CVS file version.
     *
     * @return a string indicating the inference code version.
     */
    virtual std::string GetVersion() const = 0;

    /**
     * Initialize a new instance to use the given forward model
     * and extract additional configuration from the given
     * arguments.
     * @param fwd_model Forward model to be used.
     * @param args Additional configuration parameters.
     */
    virtual void Initialize(FwdModel *fwd_model, FabberRunData &rundata);

    /**
     * Perform inference using the given model upon the given data.
     *
     * This method should only be called after Initialize()
     * Subclasses of InferenceTechnique must implement this method
     * to carry out their given inference calculations
     *
     * @param data
     */
    virtual void DoCalculations(FabberRunData &rundata) = 0;

    /**
     * Save the results
     */
    virtual void SaveResults(FabberRunData &rundata) const;

    /**
     * Destructor.
     */
    virtual ~InferenceTechnique();

protected:
    void InitMVNFromFile(FabberRunData &rundata, std::string paramFilename);

    /**
     * Pointer to forward model, passed in to initialize.
     *
     * Will not be deleted, that is the responsibility of
     * the caller
     */
    FwdModel *m_model;

    /**
     * Number of model parameters.
     *
     * This is used regularly so it's sensible to keep a
     * copy around
     */
    int m_num_params;

    /**
     * If true, stop if we get a numerical execption in any voxel. If false,
     * simply print a warning and continue
     */
    bool m_halt_bad_voxel;

    /**
     * Results of the inference method
     *
     * Vector of MVNDist, one for each voxel
     * Each MVNDist contains the means and covariance/precisions for
     * the parameters in the model
     */
    std::vector<MVNDist *> resultMVNs;

    /**
     * List of masked timepoints
     *
     * Masked timepoints are indexed starting at 1 and are ignored
     * in the analysis and parameter updates.
     */
    std::vector<int> m_masked_tpoints;

    /**
     * Include very verbose debugging output
     */
    bool m_debug;

private:
    /**
     * Private to prevent assignment
     */
    const InferenceTechnique &operator=(const InferenceTechnique &from)
    {
        assert(false);
        return from;
    }
};

/**
 * \ref SingletonFactory that returns pointers to
 * \ref InferenceTechnique.
 */
typedef SingletonFactory<InferenceTechnique> InferenceTechniqueFactory;
