/*  noisemodel.h - Class declaration for generic noise models

 Adrian Groves & Michael Chappell, FMRIB Image Analysis Group & IBME QuBIc Group

 Copyright (C) 2007-2015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#pragma once

#include "dist_mvn.h"
#include "factories.h"
#include "fwdmodel_linear.h"
#include "rundata.h"

#include "armawrap/newmat.h"

#include <ostream>
#include <string>

/**
 * Base class for parameters to a noise model
 *
 * Each derived NoiseModel will have a derived NoiseParams
 */
class NoiseParams
{
public:
    /** Virtual destructor for subclasses */
    virtual ~NoiseParams()
    {
    }
    /**
     * Create a copy of the params
     *
     * Polymorphic so subclasses can create copies of the correct class
     */
    virtual NoiseParams *Clone() const = 0;

    /** Assignment operator, needs to check subclass equivalence */
    virtual const NoiseParams &operator=(const NoiseParams &in) = 0;

    /**
     * Output as a multivariate normal dist. Noise models need not
     * be normal, but they will generally have means and precisions
     */
    virtual const MVNDist OutputAsMVN() const = 0;

    /**
     * Input from MVN distribution - see OutputAsMVN
     */
    virtual void InputFromMVN(const MVNDist &mvn) = 0;

    /**
     * Dump human-readable debug output to output stream
     */
    virtual void Dump(std::ostream &os) const = 0;
};

/**
 * Generic noise model
 *
 * This class & derived classes should be essentially data-free, instead storing
 * the relevant noise parameters in a NoiseParams-derived subclass
 *
 * FIXME unclear why params need to be separate?
 */
class NoiseModel : public Loggable
{
public:
    /**
     * Static member function, to pick a noise model from a name
     */
    static NoiseModel *NewFromName(const std::string &name);

    /**
     * Create a new instance of this class. Subclasses
     * implement this to produce an instance of themselves
     *
     * @return pointer to new instance.
     */
    static NoiseModel *NewInstance();

    /** Virtual destrictor for subclasses */
    virtual ~NoiseModel()
    {
    }
    /**
     * Initialize a new instance using configuration from the given
     * arguments.
     *
     * @param args Configuration parameters.
     */
    virtual void Initialize(FabberRunData &args);

    /**
     * Create a new instance of the noise parameters
     */
    virtual NoiseParams *NewParams() const = 0;

    /**
     * Suggest some nice default values for noise parameters:
     */
    virtual void HardcodedInitialDists(NoiseParams &prior, NoiseParams &posterior) const = 0;

    /**
     * Some noise models might want to precalculate things (for efficiency
     * reasons), based on the length of the data... if you don't know what
     * this is for then just ignore it.
     */
    virtual void Precalculate(NoiseParams &noise, const NoiseParams &noisePrior,
        const NEWMAT::ColumnVector &sampleData) const
    {
    }

    // VB Updates
    // The following could potentially be split into substeps; but since
    // these would necessarily be model-specific, it's nice to have a
    // general catch-all update step.  Presumably this function
    // would call all the other functions in some order.

    /**
     * Update noise parameters
     */
    virtual void UpdateNoise(NoiseParams &noise, const NoiseParams &noisePrior,
        const MVNDist &theta, const LinearFwdModel &model,
        const NEWMAT::ColumnVector &data) const = 0;

    /**
     * Update model parameters
     *
     * @param thetaWithoutPrior used for --spatial-prior-output-correction
     */
    virtual void UpdateTheta(const NoiseParams &noise, MVNDist &theta, const MVNDist &thetaPrior,
        const LinearFwdModel &model, const NEWMAT::ColumnVector &data,
        MVNDist *thetaWithoutPrior = NULL, float LMalpha = 0) const = 0;

    /**
     * Calculate the free energy
     *
     * This may be used to assess convergence
     */
    virtual double CalcFreeEnergy(const NoiseParams &noise, const NoiseParams &noisePrior,
        const MVNDist &theta, const MVNDist &thetaPrior, const LinearFwdModel &model,
        const NEWMAT::ColumnVector &data) const = 0;

    /**
     * Initialize ARD
     */
    double SetupARD(std::vector<int> ardindices, const MVNDist &theta, MVNDist &thetaPrior) const;

    /**
     * Update ARD
     */
    double UpdateARD(std::vector<int> ardindices, const MVNDist &theta, MVNDist &thetaPrior) const;

    /**
     * Return the number of noise parameters
     */
    virtual int NumParams() = 0;

protected:
    /**
     * List of masked timepoints
     *
     * Masked timepoints are indexed starting at 1 and are ignored
     * in the analysis and parameter updates.
     */
    std::vector<int> m_masked_tpoints;

private:
    /**
     * Prevent copying using anything other than the Clone() function.
     * Could implement it, but not particularly useful and the default
     * shallow copy is not right.
     */
    const NoiseModel &operator=(const NoiseModel &) const
    {
        assert(false);
        return *this;
    }
};

/**
 * \ref SingletonFactory that returns pointers to \ref NoiseModel.
 */
typedef SingletonFactory<NoiseModel> NoiseModelFactory;
