/*  fwdmodel_linear.cc - Linear forward model and related classes

 Adrian Groves & Michael Chappell, FMRIB Image Analysis Group & IBME QuBIc Group

 Copyright (C) 2007-2015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include "fwdmodel_linear.h"

#include "dist_mvn.h"
#include "easylog.h"
#include "rundata.h"
#include "tools.h"
#include "version.h"

#include "armawrap/newmat.h"

#include <iostream>
#include <stdexcept>
#include <string>
#include <vector>

using fabber::read_matrix_file;
using namespace std;
using namespace NEWMAT;

FwdModel *LinearFwdModel::NewInstance()
{
    return new LinearFwdModel();
}

static OptionSpec OPTIONS[] = { { "basis", OPT_MATRIX, "Design matrix", OPT_REQ, "" }, { "" } };

void LinearFwdModel::GetOptions(std::vector<OptionSpec> &opts) const
{
    for (int i = 0; OPTIONS[i].name != ""; i++)
    {
        opts.push_back(OPTIONS[i]);
    }
}

std::string LinearFwdModel::GetDescription() const
{
    return "Model in which output is a linear combination of input parameters";
}

string LinearFwdModel::ModelVersion() const
{
    return fabber_version();
}

void LinearFwdModel::Initialize(FabberRunData &args)
{
    FwdModel::Initialize(args);
    string designFile = args.GetString("basis");
    LOG << "LinearFwdModel::Reading design file: " << designFile << endl;
    m_jacobian = read_matrix_file(designFile);

    const int Ntimes = m_jacobian.Nrows();
    const int Nbasis = m_jacobian.Ncols();

    LOG << "LinearFwdModel::Loaded " << m_jacobian.Ncols() << " basis functions of length "
        << Ntimes << endl;

    m_centre.ReSize(Nbasis);
    m_centre = 0;
    m_offset.ReSize(Ntimes);
    m_offset = 0;

    if (args.GetBool("add-ones-regressor"))
    {
        // Add an additional 'parameter' whose timeseries if constant
        // NB: Nbasis will no longer be the number of parameters
        LOG << "LinearFwdModel::Plus an additional regressor of all ones\n";
        ColumnVector ones(Ntimes);
        ones = 1.0;
        m_jacobian = m_jacobian | ones;
        m_centre.ReSize(Nbasis + 1);
    }
}

void LinearFwdModel::GetParameterDefaults(std::vector<Parameter> &params) const
{
    for (int i = 0; i < m_centre.Nrows(); i++)
    {
        params.push_back(Parameter(
            i, "Parameter_" + stringify(i + 1), DistParams(0, 1e12), DistParams(0, 1e12)));
    }
}

void LinearFwdModel::EvaluateModel(
    const ColumnVector &params, ColumnVector &result, const std::string &key) const
{
    result = m_jacobian * (params - m_centre) + m_offset;
}

ReturnMatrix LinearFwdModel::Jacobian() const
{
    return m_jacobian;
}

ReturnMatrix LinearFwdModel::Centre() const
{
    return m_centre;
}

ReturnMatrix LinearFwdModel::Offset() const
{
    return m_offset;
}

LinearizedFwdModel::LinearizedFwdModel(const FwdModel *model)
    : m_model(model)
{
    SetLogger(model->GetLogger());
}

LinearizedFwdModel::LinearizedFwdModel(const LinearizedFwdModel &from)
    : LinearFwdModel(from)
    , m_model(from.m_model)
{
    SetLogger(from.GetLogger());
}

void LinearizedFwdModel::ReCentre(const ColumnVector &about)
{
    assert(about == about); // isfinite

    // Store new centre & offset
    m_centre = about;

    m_model->EvaluateFabber(m_centre, m_offset);
    if (0 * m_offset != 0 * m_offset)
    {
        LOG_ERR("LinearizedFwdModel::about:\n" << about);
        LOG_ERR("LinearizedFwdModel::m_offset:\n" << m_offset.t());
        throw FabberInternalError(
            "LinearizedFwdModel::ReCentre: Non-finite values found in offset");
    }

    // Calculate the Jacobian numerically.  jacobian is len(y)-by-len(m)
    m_jacobian.ReSize(m_offset.Nrows(), m_centre.Nrows());

    // Try and get the gradient matrix (Jacobian) from the model first
    // FIXME this is broken when transforms are used
    // int gradfrommodel = m_model->Gradient(m_centre, m_jacobian);

    // If the gradient is not supported by the model, use
    // numerical differentiation to calculate it.
    if (true)
    {
        ColumnVector centre2, centre3;
        ColumnVector offset2, offset3;
        for (int i = 1; i <= m_centre.Nrows(); i++)
        {
            double delta = m_centre(i) * 1e-5;
            if (delta < 0)
                delta = -delta;
            if (delta < 1e-10)
                delta = 1e-10;

            // Take derivative numerically
            centre3 = m_centre;
            centre2 = m_centre;
            centre2(i) += delta;
            centre3(i) -= delta;
            m_model->EvaluateFabber(centre2, offset2);
            m_model->EvaluateFabber(centre3, offset3);
            m_jacobian.Column(i) = (offset2 - offset3) / (centre2(i) - centre3(i));
        }
    }

    if (0 * m_jacobian != 0 * m_jacobian)
    {
        LOG << "LinearizedFwdModel::jacobian:\n" << m_jacobian;
        LOG << "LinearizedFwdModel::about':\n" << about.t();
        LOG << "LinearizedFwdModel::offset':\n" << m_offset.t();
        throw FabberInternalError(
            "LinearizedFwdModel::ReCentre: Non-finite values found in jacobian");
    }
}
