/*  fwdmodel_linear.h - Linear forward model and related classes

 Adrian Groves & Michael Chappell, FMRIB Image Analysis Group & IBME QuBIc Group

 Copyright (C) 2007-20015 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#pragma once

#include "fwdmodel.h"

#include "armawrap/newmat.h"

#include <string>
#include <vector>

/**
 * Linear forward model
 *
 * This is a model in which the output depends linearly on
 * the parameters
 *
 * The model is defined by a design matrix whose columns are each
 * a timeseries for a single parameter. The number of rows must
 * therefore equal the number of timeseries samples.
 */
class LinearFwdModel : public FwdModel
{
public:
    static FwdModel *NewInstance();

    virtual void GetOptions(std::vector<OptionSpec> &opts) const;
    virtual std::string GetDescription() const;
    virtual std::string ModelVersion() const;

    virtual void Initialize(FabberRunData &args);

    /**
     * Evaluate the model.
     *
     * In the simplest case the model is evaluated by taking the product of the
     * design
     * matrix with the parameter vector. However the model can be recentred and
     * offsetted
     * so with this in mind the evaluation consists of:
     *
     * R = J x (P - C) + O
     *
     * Where R is the output result, J is the Jacobian or design matrix, P are the
     * parameters,
     * C is the centre and O is the offset.
     *
     * Centre and offset will be zero for a basic linear model, however a subclass
     * can set
     * these to non-zero values, e.g. in LinearizedFwdModel
     *
     * @param params Parameter vector, P in above equation. Length equal to number
     * of parameters
     *               i.e. number of columns in design matrix.
     * @param result Result vector, R in above equation. Length equal to number of
     * time samples
     *               i.e. number of rows in design matrix.
     */
    virtual void EvaluateModel(const NEWMAT::ColumnVector &params, NEWMAT::ColumnVector &result,
        const std::string &key = "") const;

    /**
     * @return the Jacobian, or design matrix
     */
    NEWMAT::ReturnMatrix Jacobian() const;

    /**
     * @return the vector used to recentre the parameters
     */
    NEWMAT::ReturnMatrix Centre() const;

    /**
     * @return the vector used to offset the result vector
     */
    NEWMAT::ReturnMatrix Offset() const;

protected:
    virtual void GetParameterDefaults(std::vector<Parameter> &params) const;

    NEWMAT::Matrix m_jacobian;     // J (tranposed?)
    NEWMAT::ColumnVector m_centre; // m
    NEWMAT::ColumnVector m_offset; // g(m) - The amount to effectively subtract from Y is g(m)-J*m
};

/**
 * Linearized wrapper interface to another nonlinear forward model
 *
 * This is not used as a model directly so we do not need to implement
 * parameter defaults, version numbers, options, etc. It is debatable whether
 * it should be a FwdModel instance at all.
 */
class LinearizedFwdModel : public LinearFwdModel
{
public:
    /**
     * Create a linear approximation to another (nonlinear) forward model
     *
     * This model implements a linear approximation to the given
     * model. The ReCentre method must be called to update the
     * Jacobian, offset and centre so that Evaluate returns
     * the linear approximation which is correct at the given
     * centre.
     *
     * The model pointer is not owned by this class and will not be freed
     */
    explicit LinearizedFwdModel(const FwdModel *model);

    /**
     * Copy constructor (needed for using vector<LinearizedFwdModel>)
     *
     * NOTE: This is a reference, not a pointer... and it *copies* the
     * given LinearizedFwdModel, rather than using it as its nonlinear model!
     */
    LinearizedFwdModel(const LinearizedFwdModel &from);

    /**
     * Re-calculate the linearized model about the given centre
     *
     * @param about New centre in parameter space
     *
     * This function sets the model centre to the vector given
     * and recalculates the Jacobian and offset to form a linear
     * approximation to the underlying model.
     *
     * The offset is set to the evaluation of the underlying
     * model on the new centre, to ensure that Evaluate is
     * correct when called with the given centre
     *
     * The Jacobian is calculated either by asking the model
     * directly using the \ref Gradient method, or if that is not
     * implemented by numerical differentiation about the
     * new centre
     */
    void ReCentre(const NEWMAT::ColumnVector &about);

private:
    const FwdModel *m_model;
};
