#pragma once
/**
 * transforms.h
 *
 * Classes defining transformations on model parameters. These may be used
 * as a means of preventing model parameters from violating hard limits,
 * such as not becoming negative. Transformations are assumed to be
 * stateless, hence singleton instances can be used.
 *
 * Copyright (C) 2007-2017 University of Oxford
 */

/*  Part of FSL - FMRIB's Software Library
    http://www.fmrib.ox.ac.uk/fsl
    fsl@fmrib.ox.ac.uk

    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK


    LICENCE

    FMRIB Software Library, Release 6.0 (c) 2018, The University of
    Oxford (the "Software")

    The Software remains the property of the Oxford University Innovation
    ("the University").

    The Software is distributed "AS IS" under this Licence solely for
    non-commercial use in the hope that it will be useful, but in order
    that the University as a charitable foundation protects its assets for
    the benefit of its educational and research purposes, the University
    makes clear that no condition is made or to be implied, nor is any
    warranty given or to be implied, as to the accuracy of the Software,
    or that it will be suitable for any particular purpose or for use
    under any specific conditions. Furthermore, the University disclaims
    all responsibility for the use which is made of the Software. It
    further disclaims any liability for the outcomes arising from using
    the Software.

    The Licensee agrees to indemnify the University and hold the
    University harmless from and against any and all claims, damages and
    liabilities asserted by third parties (including claims for
    negligence) which arise directly or indirectly from the use of the
    Software or the sale of any products based on the Software.

    No part of the Software may be reproduced, modified, transmitted or
    transferred in any form or by any means, electronic or mechanical,
    without the express permission of the University. The permission of
    the University is not required if the said reproduction, modification,
    transmission or transference is done without financial return, the
    conditions of this Licence are imposed upon the receiver of the
    product, and all original and amended source code is included in any
    transmitted product. You may be held legally responsible for any
    copyright infringement that is caused or encouraged by your failure to
    abide by these terms and conditions.

    You are not permitted under this Licence to use this Software
    commercially. Use for which any financial return is received shall be
    defined as commercial use, and includes (1) integration of all or part
    of the source code or the Software into a product for sale or license
    by or on behalf of Licensee to third parties or (2) use of the
    Software or any derivative of it for research with the final aim of
    developing software products for sale or license to a third party or
    (3) use of the Software or any derivative of it for research with the
    final aim of developing non-software products for sale or license to a
    third party, or (4) use of the Software to provide any service to an
    external organisation for which payment is received. If you are
    interested in using the Software commercially, please contact Oxford
    University Innovation ("OUI"), the technology transfer company of the
    University, to negotiate a licence. Contact details are:
    fsl@innovation.ox.ac.uk quoting Reference Project 9564, FSL.*/

#include "rundata.h"

#include <math.h>
#include <string>

const std::string TRANSFORM_CODE_IDENTITY = "I";
const std::string TRANSFORM_CODE_LOG = "L";
const std::string TRANSFORM_CODE_SOFTPLUS = "S";
const std::string TRANSFORM_CODE_FRACTIONAL = "F";
const std::string TRANSFORM_CODE_ABS = "A";

/**
 * Immutable object describing distribution parameters for single model parameter
 */
struct DistParams
{
    DistParams(double m = 0, double v = 1)
        : m_mean(m)
        , m_var(v)
        , m_prec(1 / v)
    {
    }
    double mean() const
    {
        return m_mean;
    }
    double var() const
    {
        return m_var;
    }
    double prec() const
    {
        return m_prec;
    }

private:
    double m_mean;
    double m_var;
    double m_prec;
};

/**
 * Abstract base class for parameter transformations.
 */
class Transform
{
public:
    virtual ~Transform()
    {
    }

    /**
     * Transform the Fabber internal value (which is assumed to have a Gaussian
     * distribution) to the value required by the model
     */
    virtual double ToModel(double val) const = 0;

    /**
     * Transform a model parameter value to the value to be used by Fabber internally
     * (and modelled as a Gaussian random variable)
     */
    virtual double ToFabber(double val) const = 0;

    /**
     * Transform the Fabber internal variance (which is assumed to have a Gaussian
     * distribution) to the value required by the model
     */
    virtual double ToModelVar(double val) const;

    /**
     * Transform the model variance to the variance of the Gaussian to be used by Fabber
     * internally
     */
    virtual double ToFabberVar(double val) const;

    /**
     * Transform the Fabber internal mean/variance (of the Gaussian
     * distribution) to the mean/variance of the model distribution

     * Particular transformations may implement this differently, e.g.
     * the log transformation can use the known expression for the
     * variance of the log-normal distribution in terms of the
     * underlying normal distribution.
     */
    DistParams ToModel(DistParams params) const;

    /**
     * Transform the mean of the model distribution into the
     * mean of the corresponding Gaussian distribution.
     *
     * The default just applies ToFabber to the mean and transforms
     * the variance in an ad-hoc manner - @see ToModel for details.
     */
    DistParams ToFabber(DistParams params) const;
};

/**
 * Trivial implementation which performs no transformation
 */
class IdentityTransform : public Transform
{
public:
    double ToModel(double val) const
    {
        return val;
    }
    double ToFabber(double val) const
    {
        return val;
    }
    double ToModelVar(double val) const
    {
        return val;
    }
    double ToFabberVar(double val) const
    {
        return val;
    }
};

/**
 * Transformation for a parameter which is log-normal
 */
class LogTransform : public Transform
{
public:
    double ToModel(double val) const
    {
        return exp(val);
    }
    double ToFabber(double val) const
    {
        return log(val);
    }
    double ToModelVar(double val) const
    {
        return exp(val);
    }
    double ToFabberVar(double val) const
    {
        return log(val);
    }
};

/**
 * 'Softplus' transformation
 *
 * This is an alternative to the log transform for a parameter which
 * is strictly positive. For positive values it approaches the
 * identity transformation so avoiding any issues associated with the
 * rapid growth of the exponential function.
 */
class SoftPlusTransform : public Transform
{
public:
    double ToModel(double val) const
    {
        if (val < 10)
        {
            return log(1 + exp(val));
        }
        else
        {
            return val;
        }
    }
    double ToFabber(double val) const
    {
        if (val < 10)
        {
            return log(exp(val) - 1);
        }
        else
        {
            return val;
        }
    }
};

/**
 * 'Fractional' transformation
 *
 * This enforces parameter values between 0 and 1, exclusive
 * The variance is not transformed at all. This transformation
 * is intended for parameters which are not biophysical and
 * therefore we are not interested in specifying informative
 * priors, or analysing output variance.
 */
class FractionalTransform : public Transform
{
public:
    double ToModel(double val) const
    {
        return 1 / (1 + exp(val));
    }
    double ToFabber(double val) const
    {
        return log(1 / val - 1);
    }
    double ToModelVar(double val) const
    {
        return val;
    }
    double ToFabberVar(double val) const
    {
        return val;
    }
};

/**
 * 'Absolute' transformation
 *
 * This enforces parameter values >= 0 using a modulus
 * function. Note that this is not invertible - we take
 * the positive branch when doing model->Fabber conversion
 */
class AbsTransform : public Transform
{
public:
    double ToModel(double val) const
    {
        return fabs(val);
    }
    double ToFabber(double val) const
    {
        return val;
    }
};

/** Singleton instance of identity transform */
const Transform *TRANSFORM_IDENTITY();

/** Singleton instance of log transform */
const Transform *TRANSFORM_LOG();

/** Singleton instance of softplus transform */
const Transform *TRANSFORM_SOFTPLUS();

/** Singleton instance of fractional transform */
const Transform *TRANSFORM_FRACTIONAL();

/** Singleton instance of fractional transform */
const Transform *TRANSFORM_ABS();

const Transform *GetTransform(std::string id);
